//Ethan Alexander Shulman 2024

var mapPoints = [], mapMines = [], batchPoints = [], mapPaths = [], mapFactionSpawns = [],
mapMineRecount = [],
mapPointsBatch, mapMinesBatch, mapPathsBatch;

const MAP_SIZES = ["Small","Medium","Large"],
SPAWN_MINE_SIZE = 15;


function InitializeMap() {
	mapPointsBatch = new SpriteBatch();
	mapMinesBatch = new SpriteBatch();
	mapPathsBatch = new SpriteBatch();
	mapPathsBatch.lines = true;
}


/**@constructor
@extends {Sprite}*/
function MapPoint(x,y) {
	Sprite.call(this);
	this.x = x;
	this.y = y;
		
	this.pointId = mapPoints.length;
	this.paths = [];
	this.connections = [];
	
	this.mineSize = 0;
	this.mineFaction = -1;
	this.spawnDelay = 0;
	this.spawnTimer = 0;
	this.factionCount = new Array(FACTION_COUNT);//count of pixels from each faction at point
	this.factionCount.fill(0);
	this.countChanged = false;
	this.travelCount = new Array(FACTION_COUNT);//count of pixels from each faction travelling to point
	this.travelCount.fill(0);
	this.selectedCount = 0;
	
	this.findPath = null;
	this.pfLengths = null;
	this.pfSkip = null;
}
MapPoint.prototype = Object.create(Sprite.prototype);



/**@constructor
@extends {Sprite}*/
function MapPath(p1,p2) {
	Sprite.call(this);
	this.pointA = p1;
	this.pointB = p2;
	this.x = (p1.x+p2.x)*0.5;
	this.y = (p1.y+p2.y)*0.5;
	
	var dx = p2.x-p1.x, dy = p2.y-p1.y;
	this.scaleX = 5;
	this.distance = Math.sqrt(dx*dx+dy*dy);
	this.scaleY = this.distance*0.5;
	this.angle = Math.atan2(dx,dy);
	this.directionX = dx/this.distance;
	this.directionY = dy/this.distance;
	
	p1.paths.push(this);
	p1.connections.push(p2);
	p2.paths.push(this);
	p2.connections.push(p1);
}
MapPath.prototype = Object.create(Sprite.prototype);


//clear map
function ClearMap() {
	mapPaths.length = 0;
	mapPathsBatch.renderCount = 0;
	mapPoints.length = 0;
	mapPointsBatch.renderCount = 0;
	batchPoints.length = 0;
	mapMines.length = 0;
	mapMinesBatch.renderCount = 0;
	mapFactionSpawns.length = 0;
	
	if (deathFxQuality) {
		deathFxA.Bind();
		gl.clearColor(0,0,0,0);
		gl.clear(gl.COLOR_BUFFER_BIT);
	}
}



//generate grid map section
function GenerateGridSection(prng,ox,oy,smallMap) {
	//points
	var section = [], baseId = mapPoints.length;
	for (var y = 0; y < 4; y++) {
		var yl = Math.abs(y-1.5);
		for (var x = 0; x < 4; x++) {
			var xl = Math.abs(x-1.5),
				pt = new MapPoint(ox-192+x*128+prng.Int(-35,36),
								oy-192+y*128+prng.Int(-35,36));				
			if (smallMap && Math.max(xl,yl) > 1.2) pt.mineSize = prng.Int(15,31);
			mapPoints.push(pt);
			section.push(pt);
		}
	}
	
	//paths
	for (var y = 0; y < 4; y++) {
		var yMid = (Math.abs(y-1.5) < 0.6);
		for (var x = 0; x < 4; x++) {
			var xMid = (Math.abs(x-1.5) < 0.6),
				id = x+y*4, nId = 0;
			if (xMid && yMid) {
				var lId = (x-1)+(y-1)*2;
				switch (lId) {
					case 0:nId = 1;break;
					case 1:nId = 4;break;
					case 2:nId = -4;break;
					case 3:nId = -1;break;
				}
				nId += id;
				
			} else {
				var dX = x-1.5, dY = y-1.5;
				nId = 5+(dX<0?0:1)+(dY<0?0:4);
			}
			var path = new MapPath(mapPoints[baseId+id],mapPoints[baseId+nId]);
			mapPaths.push(path);
		}
	}
	
	return section;
}


//generate orbital map section
function GenerateOrbitalSection(prng,ox,oy) {
	var center = new MapPoint(ox,oy),
		section = [];
	mapPoints.push(center);
	for (var i = 0; i < 15; i++) {
		var ang = i*Math.PI*2/15,
			p = new MapPoint(ox+Math.round(Math.cos(ang)*200)+prng.Int(-30,31),
							oy+Math.round(Math.sin(ang)*200)+prng.Int(-30,31));

		mapPoints.push(p);
		section.push(p);
		mapPaths.push(new MapPath(center,p));
	}
	for (var i = 0; i < 15; i++) {
		var p = section[i], np = section[(i+1)%15],
			dx = p.x-np.x, dy = p.y-np.y,
			len = Math.sqrt(dx*dx+dy*dy);
		if (len < 80) {
			mapPaths.push(new MapPath(p,np));
		}
	}
	section.push(center);
	return section;
}


//generate ring section
function GenerateRingSection(prng,ox,oy) {
	var section = [];
	for (var i = 0; i < 16; i++) {
		var ang = i*Math.PI*2/16,
			len = (i%2 ? 150 : 210),
			p = new MapPoint(ox+Math.round(Math.cos(ang)*len)+prng.Int(-30,31),
							oy+Math.round(Math.sin(ang)*len)+prng.Int(-30,31));
		mapPoints.push(p);
		section.push(p);
	}
	for (var i = 0; i < 16; i++) {
		mapPaths.push(new MapPath(section[i],section[(i+1)%16]));
	}
	var cross = prng.Int(-16,16);
	if (cross > -1) {
		mapPaths.push(new MapPath(section[cross],section[(cross+8)%16]));
	}
	return section;
}


//connect map sections by paths between nearest points
function ConnectSections(sa,sb) {
	var nDst = 1e8, nA = 0, nB = 0;
	for (var i = 0; i < 16; i++) {
		var pa = sa[i];
		for (var k = 0; k < 16; k++) {
			var pb = sb[k],
				dx = pa.x-pb.x, dy = pa.y-pb.y,
				dst = dx*dx+dy*dy;
			if (dst < nDst) {
				nDst = dst;
				nA = i;
				nB = k;
			}
		}
	}
	mapPaths.push(new MapPath(sa[nA],sb[nB]));
}


//find nearest map point to position
function NearestMapPoint(x,y) {
	var nPt = mapPoints.length, minDst = 1e16, nearest = null;
	for (var i = 0; i < nPt; i++) {
		var pt = mapPoints[i],
			dx = x-pt.x,
			dy = y-pt.y,
			len = Math.sqrt(dx*dx+dy*dy);
		if (len < minDst) {
			minDst = len;
			nearest = pt;
		}
	}
	returnY = minDst;
	return nearest;
}



//generate new map
function GenerateMap(mapSize, mapSeed) {
	/*small = 1 section
	medium = 2x2 sections
	large = 3x3
	section = 16 points across 512x512*/
	var prng = new PRNG(mapSeed);
	if (mapSize === 0) {
		//small
		GenerateGridSection(prng,1024,1024,true);
		
	} else {
		//medium/large
		var sections = [],
			gridStep = 3-mapSize, gridStride = mapSize*256,
			gridSize = mapSize+1;
		//create sections
		for (var gy = -1; gy < 2; gy += gridStep) {
			var secY = 1024+gy*gridStride;
			for (var gx = -1; gx < 2; gx += gridStep) {
				var secId = prng.Int(0,gridSize), sec = null,
					secX = 1024+gx*gridStride;
				if (secId === 0) sec = GenerateGridSection(prng,secX,secY,false);
				else if (secId === 1) sec = GenerateOrbitalSection(prng,secX,secY);
				else sec = GenerateRingSection(prng,secX,secY);
				sections.push(sec);
			}
		}
		//connect sections
		var secId = 0;
		for (var gy = 0; gy < gridSize; gy++) {
			var ny = gy+1, yOk = (ny!==gridSize);
			for (var gx = 0; gx < gridSize; gx++) {
				var nx = gx+1, xOk = (nx!==gridSize),
					sec = sections[secId];
				if (xOk) ConnectSections(sec,sections[secId+1]);
				if (yOk) ConnectSections(sec,sections[secId+gridSize]); 
				secId++;
			}
		}
	}
	
	//init faction spawns
	var facLen = activeFactions.length,
		angPerFac = Math.PI*2/facLen,
		angOffset = prng.Int(0,1000)/1000*Math.PI*2,
		spawnRadius = (mapSize+1)*280;
	for (var i = 0; i < facLen; i++) {
		var ang = angOffset+i*angPerFac,
			npoint = NearestMapPoint(1024+Math.cos(ang)*spawnRadius, 1024+Math.sin(ang)*spawnRadius);
		npoint.mineSize = SPAWN_MINE_SIZE;
		mapFactionSpawns.push(npoint);
	}
	
	//spawn mines, separate points/mines and init mines
	var mxSz = 0, mineRate = 0, mineScale = 0;
	if (mapSize == 1) {
		//medium
		mineRate = 500;
		mineScale = 0.05;
	} else if (mapSize === 2) {
		//large
		mineRate = 400;
		mineScale = 0.04;
	}
	for (var i = 0, pl = mapPoints.length; i < pl; i++) {
		var p = mapPoints[i],
			mineSz = p.mineSize;
		if (mapSize > 0 && mineSz <= 0 && prng.Int(0,1000) < mineRate) {
			//mine size based on closest distance to faction start
			var fdst = 1e16;
			for (var f = 0; f < facLen; f++) {
				var spawn = mapFactionSpawns[f],
					dx = p.x-spawn.x, dy = p.y-spawn.y;
				fdst = Math.min(fdst, Math.sqrt(dx*dx+dy*dy));
			}
			p.mineSize = mineSz = 5+Math.floor(fdst*mineScale);
			mxSz = Math.max(mxSz,mineSz);
		}
		if (mineSz > 0) {
			//init mine
			p.scaleX = p.scaleY = PIXEL_RADIUS*1.25*Math.sqrt(mineSz);
			p.spawnDelay = Math.round(500/mineSz);
			mapMines.push(p);
		
		} else {
			//init point
			p.scaleX = p.scaleY = 25;
			batchPoints.push(p);
		
		}
	}
	//console.log(mxSz);

	//build map sprite batches
	mapPointsBatch.Update(batchPoints,true);
	mapMinesBatch.Update(mapMines,true);
	mapPathsBatch.Update(mapPaths,true);
}



function UpdateMap() {
	//recount mine faction pixels to see who owns mine
	for (var i = 0; i < mapMineRecount.length; i++) {
		var mn = mapMineRecount[i], count = mn.factionCount, clen = count.length,
			maxCount = 0, maxI = 0;
		for (var k = 0; k < clen; k++) {
			var c = count[k];
			if (c > maxCount) {
				maxCount = c;
				maxI = k;
			}
		}
		if (maxCount < mn.mineSize) {
			mn.mineFaction = -1;
			mn.spawnTimer = 0;
			
		} else {
			mn.mineFaction = maxI;
		}
		mn.countChanged = false;
	}
	mapMineRecount.length = 0;
	
	//mine spawns
	for (var i = 0; i < mapMines.length; i++) {
		var mn = mapMines[i];
		if (mn.mineFaction > -1) {
			if (++mn.spawnTimer >= mn.spawnDelay) {
				FACTIONS[mn.mineFaction].Spawn(mn);
				mn.spawnTimer = 0;
			}
		}
	}
}
