//Ethan Alexander Shulman 2024

const NNAI_SRC_INPUT_SIZE = 7, NNAI_SRC_OUTPUT_SIZE = 1,
NNAI_DST_INPUT_SIZE = 54, NNAI_DST_OUTPUT_SIZE = 2,
NNAI_VIEW_EXTENT = 2, NNAI_VIEW_WIDTH = 1+NNAI_VIEW_EXTENT*2,
NNAI_VIEW_TOTAL_SIZE = NNAI_VIEW_WIDTH*NNAI_VIEW_WIDTH,
NNAI_VIEW_PIXEL_SIZE = 120, NNAI_VIEW_PIXEL_SCALE = 1/NNAI_VIEW_PIXEL_SIZE,
NNAI_DST_LEN_SCALE = 1/3000,
NNAI_DST_SURPLUS_SCALE = 1/45;


//shared array buffer allocation system
var sIB = new Int32Array(8), sFB = new Float32Array(sIB.buffer),
sabLastAlloc = -1, sabNextAlloc = 0,
sabGapAlloc = 0, sabLargestGap = 0;


//allocate region of sIB/sFB, returns pointer to start of region
function SAB_Alloc(count) {
	count += 3;
	var lastAlloc = -1, nextAlloc = -1,
		allocPos = 0;
	//check for space at largest gap
	if (sabLargestGap >= count) {
		if (!(sabGapAlloc === 0 && sIB[sabGapAlloc] === 0)) lastAlloc = sabGapAlloc;
		nextAlloc = sIB[sabGapAlloc+2];
		sabGapAlloc += sIB[sabGapAlloc];
		allocPos = sabGapAlloc;
		sabLargestGap -= count;
		
	} else {
		//check for space at end
		var bufSz = sIB.length;
		if (bufSz-sabNextAlloc < count) {
			//no existing space, expand buffer
			var nSz = bufSz,
				reqSz = bufSz+count;
			while (nSz < reqSz) nSz *= 2;
		
			var nIB = new Int32Array(nSz);
			nIB.set(sIB);
			sIB = nIB;
			sFB = new Float32Array(nIB.buffer);
		}
		
		lastAlloc = sabLastAlloc;
		allocPos = sabNextAlloc;
		sabLastAlloc = sabNextAlloc;
		sabNextAlloc += count;
	}
	
	if (lastAlloc > -1) {
		sIB[lastAlloc+2] = allocPos;
	}
	if (nextAlloc > -1) {
		sIB[nextAlloc+1] = allocPos;
	}
	sIB[allocPos++] = count;
	sIB[allocPos++] = lastAlloc;
	sIB[allocPos++] = nextAlloc;
	return allocPos;
}

//free allocated region of sIB/sFB
function SAB_Free(ptr) {
	ptr -= 3;
	var sz = sIB[ptr],
		last = sIB[ptr+1],
		next = sIB[ptr+2];
	if (last > -1) {
		sIB[last+2] = next;
		if (next > -1) {
			var gap = next-(last+sIB[last]);
			if (gap > sabLargestGap) {
				sabLargestGap = gap;
				sabGapAlloc = last;
			}	
		}
	}
	
	if (next > -1) {
		sIB[next+1] = last;
		if (last < 0) {
			if (next > sabLargestGap) {
				sabLargestGap = next;
				sabGapAlloc = 0;
				sIB[0] = 0;
				sIB[2] = next;
			}
		}
		
	} else {
		sabLastAlloc = last;
		if (last < 0) {
			sabLargestGap = sabNextAlloc = 0;
		} else {
			sabNextAlloc = ptr;
			if (last === sabGapAlloc) {
				sabLargestGap = 0;
			}
		}
	}
}

//copy from src to dst
function SAB_Copy(dst,src,len) {
	len += src;
	while (src < len) sIB[dst++] = sIB[src++];
}

//set sFB data to 0
function SFB_Zero(ptr,len) {
	len += ptr;
	while (ptr < len) sFB[ptr++] = 0;
}



/**@constructor*/
function NeuralLayer(inSz,outSz) {
	this.inputSize = inSz;
	this.outputSize = outSz;
	this.biases = SAB_Alloc((inSz+1)*outSz);
	this.weights = this.biases+outSz;
}

//process input forward through layer to output
NeuralLayer.prototype.Forward = function(input, output) {
	var inSz = this.inputSize, outSz = this.outputSize,
		bi = this.biases, wi = this.weights,
		inputEnd = input+inSz, outEnd = output+outSz;
	while (output < outEnd) {
		var v = 0;
		for (var i = input; i < inputEnd; i++) {
			v += sFB[i]*sFB[wi++];
		}
		sFB[output++] = v+sFB[bi++];
	}
}

//apply tanh to output
NeuralLayer.prototype.ApplyTanh = function(out) {
	var outEnd = out+this.outputSize;
	while (out < outEnd) {
		sFB[out] = Math.tanh(sFB[out]);
		out++;
	}
}

//import weights/biases from base64
NeuralLayer.prototype.Import = function(bs) {
	bs = atob(bs);
	var bi = 0,
		dp = this.biases,
		dend = dp+(this.inputSize+1)*this.outputSize;
	while (dp < dend) {
		var c0 = bs.charCodeAt(bi++),
			c1 = bs.charCodeAt(bi++),
			c2 = bs.charCodeAt(bi++),
			c3 = bs.charCodeAt(bi++);
		sIB[dp++] = (c0 | (c1 << 8) | (c2 << 16) | (c3 << 24));
	}
}


/**@constructor*/
function NeuralNetwork(layerSizes) {
	this.layerSizes = layerSizes;
	this.layers = new Array(layerSizes.length-1);
	
	var inSz = layerSizes[0];
	for (var i = 0; i < this.layers.length; i++) {
		var outSz = layerSizes[i+1];
		this.layers[i] = new NeuralLayer(inSz, outSz);
		inSz = outSz;
	}
}

//new state buffer
NeuralNetwork.prototype.NewState = function() {
	var sizes = this.layerSizes,
		states = new Array(sizes.length),
		total = 0;
	for (var i = 0; i < states.length; i++) total += sizes[i];
	
	states[0] = SAB_Alloc(total);
	var sz1 = states.length-1;
	for (var i = 0; i < sz1; i++) {
		states[i+1] = states[i]+sizes[i];
	}
	
	return states;
}

//propagate input state forward through layers to output
NeuralNetwork.prototype.Forward = function(state) {
	var inp = state[0],
		layers = this.layers,
		nLayers = layers.length,
		lastLayer = nLayers-1;
	for (var i = 0; i < nLayers; i++) {
		var out = state[i+1],
			lay = layers[i];
		lay.Forward(inp, out);
		if (i !== lastLayer) lay.ApplyTanh(out);
		inp = out;
	}
}


var nnaiPresent = false, nnaiViewOld = true,
nnaiMapPtCells = null, nnaiCachedView = -1, nnaiViewPixelScale = 1,
nnaiStateBuffer = -1, nnaiStateBufferSize = 0;

function NNAI_MaxStateSize(layers) {
	var sum = 0;
	for (var i = 0; i < layers.length; i++) sum += layers[i];
	nnaiStateBufferSize = Math.max(nnaiStateBufferSize, sum);
}

/**@constructor*/
function NNAI(srcDat,dstDat) {
	this.srcData = srcDat;
	this.dstData = dstDat;
	NNAI_MaxStateSize(srcDat[0]);
	NNAI_MaxStateSize(dstDat[0]);
	
	this.srcNet = null;
	this.dstNet = null;
	
	this.srcState = null;
	this.dstState = null;
}

//load neuralnet
function NNAI_LoadNet(dat) {
	var net = new NeuralNetwork(dat[0]);
	for (var i = 0; i < net.layers.length; i++) {
		net.layers[i].Import(dat[i+1]);
	}
	return net;
}
//init state using shared state buffer
function NNAI_LoadState(net) {
	var lays = net.layerSizes,
		state = new Array(lays.length),
		sptr = nnaiStateBuffer;
	for (var i = 0; i < lays.length; i++) {
		state[i] = sptr;
		sptr += lays[i];
	}
	return state;
}

//load ai from data array
NNAI.prototype.Load = function() {
	if (nnaiStateBuffer === -1) nnaiStateBuffer = SAB_Alloc(nnaiStateBufferSize);

	this.srcNet = NNAI_LoadNet(this.srcData);
	this.srcState = NNAI_LoadState(this.srcNet);
	this.srcData = null;
	
	this.dstNet = NNAI_LoadNet(this.dstData);
	this.dstState = NNAI_LoadState(this.dstNet);
	this.dstData = null;
}

//load map for nnai view
function NNAI_LoadMap() {
	if (!nnaiMapPtCells) {
		nnaiMapPtCells = new Array(144);
		for (var i = 0; i < 144; i++) {
			var cells = new Array(25);
			nnaiMapPtCells[i] = cells;
			for (var k = 0; k < 25; k++) cells[k] = [];
		}
		nnaiCachedView = SAB_Alloc(144*NNAI_VIEW_TOTAL_SIZE*MAX_FACTIONS);
	}
	
	//load map points into cells for nn view
	var nPts = mapPoints.length;
	for (var p = 0; p < nPts; p++) {
		var cells = nnaiMapPtCells[p], cellI = 0,
			mapPt = mapPoints[p];
		for (var vy = -NNAI_VIEW_EXTENT; vy <= NNAI_VIEW_EXTENT; vy++) {
			var cellY = mapPt.y+vy*NNAI_VIEW_PIXEL_SIZE;
			for (var vx = -NNAI_VIEW_EXTENT; vx <= NNAI_VIEW_EXTENT; vx++) {
				var cellX = mapPt.x+vx*NNAI_VIEW_PIXEL_SIZE,
					cell = cells[cellI++];
				cell.length = 0;
				for (var t = 0; t < nPts; t++) {
					var tpt = mapPoints[t],
						diffX = Math.abs(cellX-tpt.x),
						diffY = Math.abs(cellY-tpt.y);
					if (Math.max(diffX,diffY) < NNAI_VIEW_PIXEL_SIZE) {
						cell.push(t, (1-diffX*NNAI_VIEW_PIXEL_SCALE)*(1-diffY*NNAI_VIEW_PIXEL_SCALE));
					}
				}
			}
		}
	}
}

//update cached nnai view of map state
function NNAI_UpdateView() {
	var sptr = nnaiCachedView,
		nPts = mapPoints.length,
		maxPxCount = 0;
	for (var p = 0; p < nPts; p++) {
		var pt = mapPoints[p],
			pfc = pt.factionCount, ptc = pt.travelCount;
		for (var f = 0; f < MAX_FACTIONS; f++) {
			maxPxCount = Math.max(maxPxCount, pfc[f]+ptc[f]);
		}
			
		var cells = nnaiMapPtCells[p];
		for (var c = 0; c < NNAI_VIEW_TOTAL_SIZE; c++) {
			var cell = cells[c], cellLen = cell.length;
			SFB_Zero(sptr, MAX_FACTIONS);
			for (var k = 0; k < cellLen; k++) {
				var cPt = mapPoints[cell[k++]],
					cfc = cPt.factionCount, ctc = cPt.travelCount,
					weight = cell[k];
				for (var f = 0; f < MAX_FACTIONS; f++) {
					sFB[sptr+f] += (cfc[f]+ctc[f])*weight;
				}
			}
			sptr += MAX_FACTIONS;
		}
	}
	nnaiViewPixelScale = (maxPxCount ? 1/maxPxCount : 0);
	
	nnaiViewOld = false;
}



/**@constructor
@extends {AI}*/
function NeuralNetworkAI(pId,seed, nnai) {
	AI.call(this,pId,seed);	
	this.nn = nnai;
	if (!nnai.srcNet) nnai.Load();
	
	if (!nnaiPresent) {
		NNAI_LoadMap();
		nnaiPresent = true;
	}
}
NeuralNetworkAI.prototype = Object.create(AI.prototype);

NeuralNetworkAI.prototype.Decide = function() {
	if (nnaiViewOld) NNAI_UpdateView();

	//calculate max surplus
	var maxSurplus = 0,
		nPts = mapPoints.length,
		facId = this.faction;
	for (var p = 0; p < nPts; p++) {
		var pt = mapPoints[p];
		maxSurplus = Math.max(maxSurplus, pt.factionCount[facId]-pt.mineSize);
	}
	if (maxSurplus) maxSurplus = 1/maxSurplus;
	else return;//no surplus, no src
	
	//run src forward
	var srcNet = this.nn.srcNet,
		srcState = this.nn.srcState,
		srcOutLayer = srcNet.layers.length,
		srcTotal = 0,
		afLen = activeFactions.length;
	for (var p = 0; p < nPts; p++) {
		var pt = mapPoints[p],
			surplus = pt.factionCount[facId]-pt.mineSize;
		if (surplus > 0) {
			//src input
			var sptr = srcState[0],
				vptr = nnaiCachedView+p*NNAI_VIEW_TOTAL_SIZE*MAX_FACTIONS;
			SFB_Zero(sptr, 6);
			for (var gy = -NNAI_VIEW_EXTENT; gy <= NNAI_VIEW_EXTENT; gy++) {
				var agy = Math.abs(gy);
				for (var gx = -NNAI_VIEW_EXTENT; gx <= NNAI_VIEW_EXTENT; gx++) {
					var friend = 0, enemy = 0;
					for (var f = 0; f < MAX_FACTIONS; f++) {
						if (f === facId) friend += sFB[vptr++];
						else enemy += sFB[vptr++];
					}
					
					var ind = sptr+Math.max(Math.abs(gx), agy)*2;
					sFB[ind] += friend;
					sFB[ind+1] += enemy;
				}
			}
			
			var pxs = nnaiViewPixelScale;
			sFB[sptr] *= pxs;
			sFB[sptr+1] *= pxs;
			pxs /= 8;
			sFB[sptr+2] *= pxs;
			sFB[sptr+3] *= pxs;
			pxs /= 2;
			sFB[sptr+4] *= pxs;
			sFB[sptr+5] *= pxs;
			sFB[sptr+6] = surplus*maxSurplus;
			
			srcNet.Forward(srcState);
			
			//src point probability from output
			var prob = Math.exp(Math.min(30, sFB[srcState[srcOutLayer]]));
			abuf.push(prob, p);
			srcTotal += prob;
		}
	}
	if (srcTotal === 0) return;//no src

	//select src by randomly sampling probabilities
	var srcRnd = this.RandFloat(0, srcTotal),
		rndT = 0,
		srcLen = abuf.length,
		srcInd = -1;
	for (var i = 0; i < srcLen; i += 2) {
		var nxt = rndT+abuf[i];
		if (srcRnd < nxt) {
			srcInd = abuf[i+1];
			break;
		}
		rndT = nxt;
	}
	if (srcInd === -1) srcInd = abuf[srcLen-1];
	abuf.length = 0;
	
	
	//run dst forward
	var srcPt = mapPoints[srcInd],
		srcSurplus = srcPt.factionCount[facId]-srcPt.mineSize,
		
		dstNet = this.nn.dstNet,
		dstState = this.nn.dstState,
		dstOutLayer = dstNet.layers.length,
		dstTotal = 0;
	for (var p = 0; p < nPts; p++) {
		if (p === srcInd) continue;
		
		//dst input
		var pt = mapPoints[p],
			dptr = dstState[0],
			vptr = nnaiCachedView+p*NNAI_VIEW_TOTAL_SIZE*MAX_FACTIONS;
		for (var v = 0; v < NNAI_VIEW_TOTAL_SIZE; v++) {
			var friend = 0, enemy = 0;
			for (var f = 0; f < MAX_FACTIONS; f++) {
				if (f === facId) friend += sFB[vptr++];
				else enemy += sFB[vptr++];
			}
			
			sFB[dptr++] = friend*nnaiViewPixelScale;
			sFB[dptr++] = enemy*nnaiViewPixelScale;
		}
		
		var ptSurplus = pt.factionCount[facId]-pt.mineSize;
		sFB[dptr++] = (ptSurplus >= 0 ? (pt.mineSize>0 ? -1 : -0.1) : -ptSurplus*NNAI_DST_SURPLUS_SCALE/srcSurplus);
		sFB[dptr++] = srcPt.pointLengths[p]*NNAI_DST_LEN_SCALE;
		var dirX = srcPt.x-pt.x,
			dirY = srcPt.y-pt.y,
			dlen = Math.sqrt(dirX*dirX+dirY*dirY);
		if (dlen) {
			dirX /= dlen;
			dirY /= dlen;
		}
		sFB[dptr++] = dirX;
		sFB[dptr++] = dirY;
		
		dstNet.Forward(dstState);
		
		//dst point probability
		dptr = dstState[dstOutLayer];
		var prob = Math.exp(Math.min(30, sFB[dptr++])),
			size = Math.max(0,Math.min(1, sFB[dptr]));
		abuf.push(prob,size);
		dstTotal += prob;
	}
	
	//select dst by randomly sampling probabilities
	var dstRnd = this.RandFloat(0, dstTotal),
		dstInd = -1, dstSize = 0, aind = 0;
	rndT = 0;
	for (var p = 0; p < nPts; p++) {
		if (p === srcInd) continue;
		
		var nxt = rndT+abuf[aind++];
		if (dstRnd < nxt) {
			dstSize = abuf[aind];
			dstInd = p;
			break;
		}
		aind++;
		rndT = nxt;
	}
	if (dstInd === -1) {
		dstInd = nPts-1;
		if (dstInd === srcInd) dstInd--;
		dstSize = abuf[abuf.length-1];
	}
	abuf.length = 0;
	
	//move pixels
	var sendCount = Math.ceil(dstSize*srcSurplus);
	if (sendCount > 0) {
		this.MovePixels(srcPt, mapPoints[dstInd], sendCount);
	}
}



//ai game recording for training neuralnets
var aiRecording = null;

/**@constructor*/
function AIRecording(mask) {
	this.mask = mask;
	this.tick = 0;
	this.states = [];
	this.moves = new Array(MAX_FACTIONS);
	for (var i = 0; i < MAX_FACTIONS; i++) this.moves[i] = [];
}

AIRecording.prototype.Clear = function() {
	this.states.length = 0;
	for (var i = 0; i < MAX_FACTIONS; i++) this.moves[i].length = 0;
	this.tick = 0;
}

AIRecording.prototype.RecordState = function() {
	var nPts = mapPoints.length,
		afl = activeFactions.length,
		state = this.states;
	for (var i = 0; i < nPts; i++) {
		var p = mapPoints[i],
			fc = p.factionCount,
			tc = p.travelCount;
		for (var k = 0; k < afl; k++) {
			var afi = activeFactions[k];
			state.push(fc[afi]);
			state.push(tc[afi]);
		}
	}
	this.tick++;
}

AIRecording.prototype.RecordMove = function(aid,src,dst,count) {
	if (!this.mask[aid]) return;
	
	var surplus = src.factionCount[activeFactions[aid]]-src.mineSize;
	if (surplus <= 0) return;
	
	var moves = this.moves[aid];
	moves.push(this.tick-1);
	moves.push(src.pointId);
	moves.push(dst.pointId);
	moves.push(Math.min(1,count/surplus));
}

AIRecording.prototype.RecordGame = function(dat) {
	var winId = dat["winner"],
		states = this.states,
		moves = this.moves[winId];
	if (moves.length === 0) {
		states = moves = null;
	}
	self.postMessage({"msg":MMSG_RECORDING, "s":states, "m":moves});
}


function StartRecording(dat) {
	if (!DEV_VERSION) return;
	if (aiRecording) return;
	aiRecording = new AIRecording(dat["mask"]);
}


//neuralnet training
if (DEV_VERSION) {

	//randomize weights and reset biases to 0
	NeuralLayer.prototype.Randomize = function(scale) {
		var inSz = this.inputSize, outSz = this.outputSize,
			bi = this.biases, wi = this.weights;
		scale /= inSz;
		for (var o = 0; o < outSz; o++) {
			sFB[bi++] = 0;
			for (var i = 0; i < inSz; i++) {
				sFB[wi++] = Math.random()*scale;
			}
		}
	}

	//process output backwards through layer to input, weights are normalized to prevent explosion
	NeuralLayer.prototype.Backward = function(output, input) {
		var inSz = this.inputSize, outSz = this.outputSize,
			wi = this.weights,
			inputEnd = input+inSz, outEnd = output+outSz;
		while (input < inputEnd) {
			var v = 0, owi = wi,
				nrm = 0;
			for (var o = output; o < outEnd; o++) {
				nrm += Math.abs(sFB[owi]);
				owi += inSz;
			}
			if (nrm) nrm = 1/nrm;
			owi = wi++;
			for (var o = output; o < outEnd; o++) {
				v += sFB[o]*sFB[owi]*nrm;
				owi += inSz;
			}
			sFB[input++] = v;
		}
	}

	//train layer
	NeuralLayer.prototype.Train = function(input, error, learningRate) {
		var inSz = this.inputSize, outSz = this.outputSize,
			bi = this.biases, wi = this.weights,
			inputEnd = input+inSz, errEnd = error+outSz;
		while (error < errEnd) {
			var err = sFB[error++]*learningRate;
			sFB[bi++] += err;
			for (var i = input; i < inputEnd; i++) {
				sFB[wi++] += sFB[i]*err;
			}
		}
	}

	//apply tanh derivative to error for back propagation
	NeuralLayer.prototype.InvertTanh = function(err, inp) {
		var inpEnd = inp+this.inputSize;
		while (inp < inpEnd) {
			var v = sFB[inp++];
			sFB[err++] *= 1-v*v;
		}
	}

	//export weights/biases as base64
	NeuralLayer.prototype.Export = function() {
		var bs = "",
			dp = this.biases,
			dend = dp+(this.inputSize+1)*this.outputSize;
		while (dp < dend) {
			bs += String.fromCharCode(sIB[dp]&0xFF);
			bs += String.fromCharCode((sIB[dp]>>>8)&0xFF);
			bs += String.fromCharCode((sIB[dp]>>>16)&0xFF);
			bs += String.fromCharCode((sIB[dp]>>>24)&0xFF);
			dp++;
		}
		return btoa(bs);
	}


	//randomize weights
	NeuralNetwork.prototype.Randomize = function(scale) {
		for (var i = 0; i < this.layers.length; i++) {
			this.layers[i].Randomize(scale);
		}
	}

	//propagate output state backward through layers
	NeuralNetwork.prototype.Backward = function(error, state) {
		//backpropagate
		var layers = this.layers,
			nLayers = layers.length,
			lastLayer = nLayers-1,
			out = error[nLayers];
		for (var i = lastLayer; i > 0; i--) {
			var inp = error[i],
				lay = layers[i];
			lay.Backward(out, inp);
			lay.InvertTanh(inp, state[i]);
			out = inp;
		}
	}

	//train network on errors
	NeuralNetwork.prototype.Train = function(state, error, learnRate) {
		var layers = this.layers,
			nLayers = layers.length;
		for (var i = 0; i < nLayers; i++) {
			layers[i].Train(state[i], error[i+1], learnRate);
		}
	}

	//export neuralnetwork as JSON/base64
	NeuralNetwork.prototype.Export = function() {
		var nns = "["+JSON.stringify(this.layerSizes),
			lays = this.layers;
		for (var i = 0; i < lays.length; i++) {
			nns += ', "'+lays[i].Export()+'"';
		}
		return nns+"]";
	}

	//copy weights/biases from nnet
	NeuralNetwork.prototype.Copy = function(nnet) {
		var lays = this.layers;
		for (var i = 0; i < lays.length; i++) {
			var lay = lays[i];
			SAB_Copy(lay.biases, nnet.layers[i].biases, (lay.inputSize+1)*lay.outputSize);
		}
	}

}


//random sample order for training neuralnet
function RandomizeOrder(buf,dst, count) {
	if (buf.length < count) buf.length = count;
	if (dst.length < count) dst.length = count;
	
	for (var i = 0; i < count; i++) buf[i] = i;
	
	while (count) {
		var rind = Math.floor(Math.random()*count);
		dst[--count] = buf[rind];
		if (rind !== count) buf[rind] = buf[count];
	}
}

function scanForNaN(ptr, len) {
	len += ptr;
	while (ptr < len) {
		if (isNaN(sFB[ptr++])) return true;
	}
	return false;
}

function TrainNeuralNet(dat) {
	if (!DEV_VERSION) return;

	//init neural net to train
	var srcLayers = dat["sh"], dstLayers = dat["dh"],
		srcLearnRate = dat["sr"], dstLearnRate = dat["dr"],
		srcRetryThreshold = dat["st"], dstRetryThreshold = dat["dt"],
		srcNoise = dat["sn"], dstNoise = dat["dn"];
	srcLayers.splice(0,0,NNAI_SRC_INPUT_SIZE);
	srcLayers.push(NNAI_SRC_OUTPUT_SIZE);
	dstLayers.splice(0,0,NNAI_DST_INPUT_SIZE);
	dstLayers.push(NNAI_DST_OUTPUT_SIZE);
	
	var srcNet = new NeuralNetwork(srcLayers),
		bestSrc = new NeuralNetwork(srcLayers),
		bestSrcLoss = 1e10,
		
		dstNet = new NeuralNetwork(dstLayers),
		bestDst = new NeuralNetwork(dstLayers),
		bestDstLoss = 1e10;
	srcNet.Randomize(dat["sw"]);
	dstNet.Randomize(dat["dw"]);
	
	var srcState = new Array(144),
		srcError = new Array(144),
		srcOutputLayer = srcLayers.length-1,
		dstState = new Array(144),
		dstError = new Array(144),
		dstOutputLayer = dstLayers.length-1;
	for (var i = 0; i < 144; i++) {
		srcState[i] = srcNet.NewState();
		srcError[i] = srcNet.NewState();
		dstState[i] = dstNet.NewState();
		dstError[i] = dstNet.NewState();
	}

	//count number of games recorded
	var recI = new Uint32Array(dat["buf"]),
		recF = new Float32Array(recI.buffer),
		gamePtrs = [], gptr = 0,
		gameOrder = [], moveOrder = [];
	while (true) {
		var nxtSize = recI[gptr];
		if (nxtSize === 0) break;
		gamePtrs.push(gptr);
		gptr += nxtSize;
	}
	console.log("Loaded "+gamePtrs.length+" games from recording.");
	
	//init image cell buffers
	var pointImageCells = new Array(144);
	for (var i = 0; i < pointImageCells.length; i++) {
		var cells = new Array(NNAI_VIEW_TOTAL_SIZE);
		pointImageCells[i] = cells;
		for (var k = 0; k < NNAI_VIEW_TOTAL_SIZE; k++) cells[k] = [];
	}
	
	var iterCount = dat["iter"];
	for (var iter = 0; iter < iterCount; iter++) {
		//iterate through games
		RandomizeOrder(abuf, gameOrder, gamePtrs.length);
		
		var srcLoss = 0, srcLossCount = 0,
			dstLoss = 0, dstLossCount = 0;
		for (var gid = 0; gid < gamePtrs.length; gid++) {
			//load game
			gptr = gamePtrs[gameOrder[gid]]+1;
			var nMapPts = recI[gptr++],
				nMoves = recI[gptr++],
				moveStride = 5+nMapPts*3,
				mapPtr = gptr+nMoves*moveStride,
				ptStride = 3+nMapPts;
				
			//load map points into image cells
			for (var p = 0; p < nMapPts; p++) {
				var cells = pointImageCells[p],	cellI = 0,
					ptIndex = mapPtr+p*ptStride,
					ptX = recF[ptIndex],
					ptY = recF[ptIndex+1];
				for (var vy = -NNAI_VIEW_EXTENT; vy <= NNAI_VIEW_EXTENT; vy++) {
					var cellY = ptY+vy*NNAI_VIEW_PIXEL_SIZE;
					for (var vx = -NNAI_VIEW_EXTENT; vx <= NNAI_VIEW_EXTENT; vx++) {
						var cellX = ptX+vx*NNAI_VIEW_PIXEL_SIZE,
							cell = cells[cellI];
						cell.length = 0;
						for (var t = 0; t < nMapPts; t++) {
							var tptIndex = mapPtr+t*ptStride,
								diffX = Math.abs(cellX-recF[tptIndex]),
								diffY = Math.abs(cellY-recF[tptIndex+1]);
							if (Math.max(diffX,diffY) < NNAI_VIEW_PIXEL_SIZE) {
								cell.push(t, (1-diffX*NNAI_VIEW_PIXEL_SCALE)*(1-diffY*NNAI_VIEW_PIXEL_SCALE));
							}
						}
						cellI++;
					}
				}
			}
			
			//iterate through moves
			RandomizeOrder(abuf, moveOrder, nMoves);
			
			for (var m = 0; m < nMoves; m++) {
				var movePtr = gptr+moveOrder[m]*moveStride,
					moveSrcId = recI[movePtr++],
					moveDstId = recI[movePtr++],
					moveSize = recF[movePtr++],
					normPixelCount = recF[movePtr++],
					normSurplus = recF[movePtr++],
					
					moveSrcPtr = mapPtr+moveSrcId*ptStride,
					moveSrcSurplus = recI[movePtr+moveSrcId*3]-recI[moveSrcPtr+2],
					moveSrcX = recF[moveSrcPtr], moveSrcY = recF[moveSrcPtr+1],
					
					dstTotal = 0, srcTotal = 0;
				if (moveSrcSurplus <= 0) {
					console.log("wtf "+moveSrcSurplus+", "+moveSrcId+", "+moveDstId+", "+moveSize+", "+m+", "+moveOrder[m]+", "+moveSrcX+", "+moveSrcY
					+", "+recI[movePtr+moveSrcId*3]+", "+recI[moveSrcPtr+2]);
					return;
				}
				
				//run forward
				for (var p = 0; p < nMapPts; p++) {
					var pPtr = mapPtr+p*ptStride,
						ptMineSz = recI[pPtr+2],
						ptSurplus = recI[movePtr+p*3]-ptMineSz,
						trainSrc = (ptSurplus > 0),
						trainDst = (p !== moveSrcId),
						dstPtState = dstState[p],
						dptr = dstPtState[0],
						srcPtState = srcState[p],
						sptr = srcPtState[0],
						imgCells = pointImageCells[p];
					if (trainSrc || trainDst) {
						//5x5 dst input image
						for (var c = 0; c < NNAI_VIEW_TOTAL_SIZE; c++) {
							var cell = imgCells[c], cellLen = cell.length,
								friend = 0, enemy = 0;
							for (var ci = 0; ci < cellLen; ci++) {
								var cptId = cell[ci++],
									cptWeight = cell[ci],
									datPtr = movePtr+cptId*3;
								friend += (recI[datPtr++]+recI[datPtr++])*cptWeight;
								enemy += recI[datPtr]*cptWeight;
							}
							sFB[dptr++] = friend*normPixelCount;
							sFB[dptr++] = enemy*normPixelCount;
						}
						if (trainDst) {
							sFB[dptr++] = (ptSurplus >= 0 ? (ptMineSz>0 ? -1 : -0.1) : -ptSurplus*NNAI_DST_SURPLUS_SCALE/moveSrcSurplus);
							sFB[dptr++] = recF[moveSrcPtr+3+p]*NNAI_DST_LEN_SCALE;
							var dirX = moveSrcX-recF[pPtr],
								dirY = moveSrcY-recF[pPtr+1],
								dlen = Math.sqrt(dirX*dirX+dirY*dirY);
							if (dlen) {
								dirX /= dlen;
								dirY /= dlen;
							}
							sFB[dptr++] = dirX;
							sFB[dptr++] = dirY;
							
							dstNet.Forward(dstPtState);
							
							dptr = dstPtState[dstOutputLayer];
							dstTotal += (sFB[dptr] = Math.exp(Math.min(30, sFB[dptr])));
							if (isNaN(dstTotal)) {
								console.log("dst total nan "+dstTotal+", "+sFB[dptr]);
								return;
							}
						}
					}
					if (trainSrc) {
						//src input
						var vptr = dstPtState[0];
						SFB_Zero(sptr, 6);
						for (var gy = -NNAI_VIEW_EXTENT; gy <= NNAI_VIEW_EXTENT; gy++) {
							var agy = Math.abs(gy);
							for (var gx = -NNAI_VIEW_EXTENT; gx <= NNAI_VIEW_EXTENT; gx++) {
								var ind = sptr+Math.max(Math.abs(gx), agy)*2;
								sFB[ind] += sFB[vptr++];
								sFB[ind+1] += sFB[vptr++];
							}
						}
						sFB[sptr+2] /= 8; sFB[sptr+3] /= 8;
						sFB[sptr+4] /= 16; sFB[sptr+5] /= 16;
						sFB[sptr+6] = ptSurplus*normSurplus;
						if (sFB[sptr+6] > 1) {
							console.log("over 1 huh "+sFB[sptr+6]);
							return;
						}
						
						srcNet.Forward(srcPtState);
						
						sptr = srcPtState[srcOutputLayer];
						srcTotal += (sFB[sptr] = Math.exp(Math.min(30, sFB[sptr])));
					}
				}
				dstTotal = 1/Math.max(1e-300, dstTotal);
				srcTotal = 1/Math.max(1e-300, srcTotal);
				
				//backpropagate errors
				for (var p = 0; p < nMapPts; p++) {
					var ptMineSz = recI[mapPtr+p*ptStride+2],
						ptSurplus = recI[movePtr+p*3]-ptMineSz,
						trainSrc = (ptSurplus > 0),
						trainDst = (p !== moveSrcId);
					
					if (trainDst) {
						//dst error
						var dstPtState = dstState[p],
							dstPtError = dstError[p],
							dptr = dstPtState[dstOutputLayer],
							eptr = dstPtError[dstOutputLayer];
						if (p === moveDstId) {
							sFB[eptr] = 1 - sFB[dptr]*dstTotal;
							sFB[eptr+1] = moveSize - Math.max(0,Math.min(1,sFB[dptr+1]));
						} else {
							sFB[eptr] = 0 - sFB[dptr]*dstTotal;
							sFB[eptr+1] = 0;
						}
						dstLoss += Math.abs(sFB[eptr])+Math.abs(sFB[eptr+1]);
						if (isNaN(dstLoss)) {
							console.log("nan dst loss, dst total: "+dstTotal+", errors: "+sFB[eptr]+", "+sFB[eptr+1]+
								", output: "+sFB[dptr]+
								", nan state: "+scanForNaN(dstPtState[0],dptr+2-dstPtState[0]));
							for (var dl = 0; dl < dstNet.layers.length; dl++) {
								var lay = dstNet.layers[dl];
								console.log("lay "+dl+" nan: "+scanForNaN(lay.biases, (lay.inputSize+1)*lay.outputSize));
							}
							return;
						}
						dstLossCount++;
						
						//inject noise into errors to get out of local minimums
						for (var r = 0; r < 2; r++) {
							var rnd = (Math.random()*2-1)*dstNoise;
							sFB[eptr+r] *= (rnd >= 0 ? 1+rnd : 1/(1-rnd));
						}
						
						dstNet.Backward(dstPtError, dstPtState);
					}
					
					if (trainSrc) {
						//src error
						var srcPtState = srcState[p],
							srcPtError = srcError[p],
							sptr = srcPtState[srcOutputLayer],
							eptr = srcPtError[srcOutputLayer];
						sFB[eptr] = (p === moveSrcId ? 1 : 0) - sFB[sptr]*srcTotal;
						srcLoss += Math.abs(sFB[eptr]);
						srcLossCount++;
						
						var rnd = (Math.random()*2-1)*srcNoise;
						sFB[eptr] *= (rnd >= 0 ? 1+rnd : 1/(1-rnd));
						
						srcNet.Backward(srcPtError, srcPtState);
					}
				}
				
				//train on errors
				for (var p = 0; p < nMapPts; p++) {
					var ptMineSz = recI[mapPtr+p*ptStride+2],
						ptSurplus = recI[movePtr+p*3]-ptMineSz,
						trainSrc = (ptSurplus > 0),
						trainDst = (p !== moveSrcId);
					
					if (trainDst) {
						dstNet.Train(dstState[p], dstError[p], dstLearnRate);
					}
					
					if (trainSrc) {
						srcNet.Train(srcState[p], srcError[p], srcLearnRate);
					}
				}
			}
		}

		srcLoss /= srcLossCount;
		if (srcLoss < bestSrcLoss) {
			bestSrc.Copy(srcNet);
			bestSrcLoss = srcLoss;
			console.log("Src loss: "+bestSrcLoss+", iteration: "+iter+".");
			
		} else if (srcLoss > bestSrcLoss*srcRetryThreshold) {
			srcNet.Copy(bestSrc);
		}
		
		dstLoss /= dstLossCount;
		if (dstLoss < bestDstLoss) {
			bestDst.Copy(dstNet);
			bestDstLoss = dstLoss;
			console.log("Dst loss: "+bestDstLoss+", iteration: "+iter+".");
			
		} else if (dstLoss > bestDstLoss*dstRetryThreshold) {
			dstNet.Copy(bestDst);
		}
	}
	
	console.log("Training complete.\nSource Neural Network("+bestSrcLoss+" loss):");
	console.log(bestSrc.Export());
	console.log("Destination Neural Network("+bestDstLoss+" loss):");
	console.log(bestDst.Export());
}

