//Ethan Alexander Shulman 2024

var pixelPool, pixelMap, deadPixels = [], selectedPixels = [], selectedBatch,
playerFactionId = 0, playerFaction = null, activeFactions = [],
pixelSepAng = 0, pixelCountTimer = 0;

const FACTION_COUNT = 6,
FACTIONS = [new Faction(0, 1,0,0), new Faction(1, 1,1,0), new Faction(2, 0,1,0), new Faction(3, 0,1,1), new Faction(4, 0,0,1), new Faction(5, 1,0,1)],

PIXELMAP_SIZE = 2048,
PIXELMAP_MAX = PIXELMAP_SIZE-1e-6,
PIXELMAP_CELL_SIZE = 8, PIXELMAP_CELL_SCALE = 1/PIXELMAP_CELL_SIZE,
PIXELMAP_CELLS = PIXELMAP_SIZE/PIXELMAP_CELL_SIZE,
PIXELMAP_CELLS_MINUS1 = PIXELMAP_CELLS-1,

PIXEL_RADIUS = 4, PIXEL_DIAMETER = PIXEL_RADIUS*2,
PIXEL_SPEED = 2,
PIXEL_ORBIT_SPEED = 0.5,
PIXEL_LIFE = 8,

PIXELMAP_CELL_BORDER = 0.5-PIXEL_DIAMETER*PIXELMAP_CELL_SCALE;


function InitializePixels() {
	pixelPool = new ObjectPool(Pixel);
	selectedBatch = new SpriteBatch();
	
	//init pixel map spatial look up for optimizing collisions
	pixelMap = new Array(PIXELMAP_CELLS*PIXELMAP_CELLS);
	for (var i = 0; i < pixelMap.length; i++) pixelMap[i] = [];
}


//pixel faction/team/color
/**@constructor*/
function Faction(id,r,g,b) {
	this.id = id;
	this.red = r;
	this.green = g;
	this.blue = b;
	this.cssColor = RGBToCSS(r,g,b);
	
	this.pixels = [];
	this.pixelsBatch = new SpriteBatch();
}

//spawn pixel at point
Faction.prototype.Spawn = function(pt) {
	var npx = pixelPool.Get();
	npx.arrayIndex = this.pixels.push(npx)-1;
	npx.faction = this.id;
	npx.collisionCount = 0;
	npx.travelPoint = null;
	npx.orbitPoint = pt;
	npx.life = PIXEL_LIFE;
	npx.selected = false;
	
	pt.factionCount[this.id]++;
	if (pt.mineSize > 0 && !pt.countChanged) {
		pt.countChanged = true;
		mapMineRecount.push(pt);
	}
	
	npx.pX = npx.lpX = pt.x;
	npx.pY = npx.lpY = pt.y;
	npx.mapIndex = Math.floor(pt.x*PIXELMAP_CELL_SCALE)+Math.floor(pt.y*PIXELMAP_CELL_SCALE)*PIXELMAP_CELLS;
	npx.cellIndex = pixelMap[npx.mapIndex].push(npx)-1;
	return npx;
}


//pixel unit
/**@constructor
@extends {Sprite}*/
function Pixel() {
	Sprite.call(this);
	this.scaleX = this.scaleY = PIXEL_RADIUS*1.25;

	this.faction = 0;
	this.arrayIndex = 0;
	this.mapIndex = -1;
	this.cellIndex = -1;
	this.selected = false;
	
	this.life = 0;
	this.pX = 0;
	this.pY = 0;
	this.lpX = 0;
	this.lpY = 0;
	this.collisionCount = 0;
	
	this.separationX = Math.cos(pixelSepAng)*0.1;
	this.separationY = Math.sin(pixelSepAng)*0.1;
	pixelSepAng = (pixelSepAng+0.01)%10000;
	
	this.orbitPoint = null;
	this.travelTarget = null;
	this.travelPoint = null;
	this.travelDistance = 0;
	this.travelEndLength = 0;
	this.travelStepX = 0;
	this.travelStepY = 0;
}
Pixel.prototype = Object.create(Sprite.prototype);


//init enemy and player faction
function SetFactions(numEnemies, colorShift) {
	activeFactions.push(playerFaction = FACTIONS[playerFactionId]);
	
	if (aiRecording) numEnemies--;
	
	for (var i = 0; i < numEnemies; i++) {
		var ci = (i+colorShift)%(FACTION_COUNT-1);
		activeFactions.push(FACTIONS[(playerFactionId+ci+1)%FACTION_COUNT]);
	}
	
	for (var i = 0; i < FACTION_COUNT; i++) {
		var ele = gameFactionCounts[i];
		ele.textContent = "";
		if (i < activeFactions.length) {
			ele.style.color = activeFactions[i].cssColor;
		}
	}
}


//clear all pixels
function ClearPixels() {
	for (var f = 0; f < FACTION_COUNT; f++) {
		var fpx = FACTIONS[f].pixels,
			nPx = fpx.length;
		for (var i = 0; i < nPx; i++) {
			var px = fpx[i];
			pixelMap[px.mapIndex].length = 0;
			pixelPool.Return(px);
		}
		fpx.length = 0;
	}
	selectedPixels.length = 0;
	activeFactions.length = 0;
	deadPixels.length = 0;
}


//send pixels travelling towards target
function PixelsTravelTarget(ppx, target) {
	var pxLen = ppx.length;
	for (var i = 0; i < pxLen; i++) {
		var px = ppx[i], fac = px.faction;
		if (px.travelPoint) {
			//already travelling, change destination
			px.travelTarget = target;
			continue;
		}
		var cpt = px.orbitPoint;
		if (cpt === target) continue;
				
		var dx = cpt.x-px.pX,
			dy = cpt.y-px.pY,
			dlen = Math.sqrt(dx*dx+dy*dy);
		px.travelTarget = target;
		px.travelPoint = cpt;
		px.travelProgress = 0;
		px.travelEndLength = dlen;
		if (dlen) dlen = PIXEL_SPEED/dlen;
		px.travelStepX = dx*dlen;
		px.travelStepY = dy*dlen;
		px.orbitPoint = null;
		
		cpt.factionCount[fac]--;
		if (cpt.mineSize > 0 && !cpt.countChanged) {
			cpt.countChanged = true;
			mapMineRecount.push(cpt);
		}
		cpt.travelCount[fac]++;
	}
}



//simulate pixels physics
function UpdatePixels() {
	//swap current position into last and update spatial look up cells
	for (var f = 0; f < activeFactions.length; f++) {
		var fpx = activeFactions[f].pixels,
			pxLen = fpx.length;
		for (var i = 0; i < pxLen; i++) {
			var px = fpx[i], x = px.pX, y = px.pY;
			px.lpX = x;
			px.lpY = y;
			
			var nMapIndex = Math.floor(x*PIXELMAP_CELL_SCALE)+Math.floor(y*PIXELMAP_CELL_SCALE)*PIXELMAP_CELLS,
				lMapIndex = px.mapIndex;
			if (nMapIndex !== lMapIndex) {
				var oldCell = pixelMap[lMapIndex],
					endInd = oldCell.length-1,
					oldCellInd = px.cellIndex;
				if (oldCellInd !== endInd) {
					var end = oldCell[endInd];
					end.cellIndex = oldCellInd;
					oldCell[oldCellInd] = end;
				}
				oldCell.length = endInd;
				
				px.mapIndex = nMapIndex;
				px.cellIndex = pixelMap[nMapIndex].push(px)-1;
			}
		}
	}

	//simulate/update
	for (var f = 0; f < activeFactions.length; f++) {
		var fpx = activeFactions[f].pixels,
			facId = activeFactions[f].id,
			pxLen = fpx.length;
		for (var i = 0; i < pxLen; i++) {
			var px = fpx[i], x = px.pX, y = px.pY,
				mX = 0, mY = 0,
				tPt = px.travelPoint, oPt = px.orbitPoint;
			if (tPt) {
				//travel to point
				mX += px.travelStepX;
				mY += px.travelStepY;
				if ((px.travelProgress += PIXEL_SPEED) >= px.travelEndLength) {
					//travelled to point
					var targ = px.travelTarget,
						backStep = (px.travelProgress-px.travelEndLength)/PIXEL_SPEED;
					mX -= px.travelStepX*backStep;
					mY -= px.travelStepY*backStep;
					if (tPt === targ) {
						//done reached final target
						px.orbitPoint = targ;
						px.travelPoint = null;
						
						targ.travelCount[facId]--;
						targ.factionCount[facId]++;
						if (targ.mineSize > 0 && !targ.countChanged) {
							targ.countChanged = true;
							mapMineRecount.push(targ);
						}
						
					} else {
						//next travel point
						var nxtId = tPt.findPath[targ.pointId],
							nxtPt = tPt.connections[nxtId],
							path = tPt.paths[nxtId],
							dscale = (path.pointB === nxtPt ? PIXEL_SPEED : -PIXEL_SPEED);
						px.travelPoint = nxtPt;
						px.travelProgress = 0;
						px.travelEndLength = path.distance;
						px.travelStepX = path.directionX*dscale;
						px.travelStepY = path.directionY*dscale;
						
						tPt.travelCount[facId]--;
						nxtPt.travelCount[facId]++;
					}
				}
				
			} else {
				//orbit point
				var oX = oPt.x-x, oY = oPt.y-y,
					oLen = Math.sqrt(oX*oX+oY*oY);
				if (oLen !== 0) {
					oLen = PIXEL_ORBIT_SPEED*(1-Math.min(5,px.collisionCount)*0.2)/oLen;
					mX += oX*oLen;
					mY += oY*oLen;
				}
			}
			
			//nearby pixel collisions
			var glX = x*PIXELMAP_CELL_SCALE, glY = y*PIXELMAP_CELL_SCALE,
				gMinX = Math.floor(glX)-1, gMinY = Math.floor(glY)-1,
				gMaxX = gMinX+3, gMaxY = gMinY+3,
				separate = false, nCollide = 0;
			gMinX = Math.max(0,gMinX); gMaxX = Math.min(PIXELMAP_CELLS,gMaxX);
			gMinY = Math.max(0,gMinY); gMaxY = Math.min(PIXELMAP_CELLS,gMaxY);
			for (var gY = gMinY; gY < gMaxY; gY++) {
				for (var gX = gMinX; gX < gMaxX; gX++) {
					var gcell = pixelMap[gX+gY*PIXELMAP_CELLS],
						cellLen = gcell.length;
					for (var c = 0; c < cellLen; c++) {
						var cPx = gcell[c];
						if (cPx === px) continue;
						var sameFaction = (cPx.faction===facId);
						if (sameFaction && (tPt || cPx.travelPoint)) continue;
						
						var dX = x-cPx.lpX, dY = y-cPx.lpY,
							dLen = Math.sqrt(dX*dX+dY*dY),
							overlap = PIXEL_DIAMETER-dLen;
						if (overlap > 0) {
							if (!sameFaction) {
								if (--px.life === 0) deadPixels.push(px);
							}
							if (dLen === 0) {
								//pixels completely overlapped, no direction to repel by so manually separate
								separate = true;
						
							} else {
								//repel pixel
								dLen = overlap*0.5/dLen;
								mX += dX*dLen;
								mY += dY*dLen;
							}
							nCollide++;
						}
					}
				}
			}
			if (separate) {
				mX += px.separationX;
				mY += px.separationY;
			}
			
			px.collisionCount = nCollide;
			px.pX = Math.max(0,Math.min(PIXELMAP_MAX, x+Math.max(-PIXEL_SPEED,Math.min(PIXEL_SPEED, mX))));
			px.pY = Math.max(0,Math.min(PIXELMAP_MAX, y+Math.max(-PIXEL_SPEED,Math.min(PIXEL_SPEED, mY))));
		}
	}
	
	var nDead = deadPixels.length;
	if (nDead) {
		if (deathFxQuality) RenderDeadPixelsFx();
	
		//destroy dead pixels
		for (var i = 0; i < nDead; i++) {
			var dpx = deadPixels[i], dfac = dpx.faction,
				tpt = dpx.travelPoint, facPx = FACTIONS[dfac].pixels,
				pxId = dpx.arrayIndex, pxEnd = facPx.length-1;
			if (pxId !== pxEnd) {
				var end = facPx[pxEnd];
				end.arrayIndex = pxId;
				facPx[pxId] = end;
			}
			facPx.length = pxEnd;
		
			var cell = pixelMap[dpx.mapIndex],
				endInd = cell.length-1,
				cellInd = dpx.cellIndex;
			if (cellInd !== endInd) {
				var end = cell[endInd];
				end.cellIndex = cellInd;
				cell[cellInd] = end;
			}
			cell.length = endInd;
		
			if (dpx.selected) {
				var sind = selectedPixels.indexOf(dpx);
				if (sind !== -1) selectedPixels.fastRemove(sind);
				
				if (!tpt) {
					dpx.orbitPoint.selectedCount--;
				}
			}
		
			if (tpt) {
				tpt.travelCount[dfac]--;
				
			} else {
				var opt = dpx.orbitPoint;
				opt.factionCount[dfac]--;
				if (opt.mineSize > 0 && !opt.countChanged) {
					opt.countChanged = true;
					mapMineRecount.push(opt);
				}
			}
			
			pixelPool.Return(dpx);
		}
		deadPixels.length = 0;
	}
	
	
	//update faction count text and check if players dead/won
	if (appStatus === APPSTATUS_GAME && ++pixelCountTimer >= 10 && !inTutorial) {
		pixelCountTimer = 0;

		if (DEV_VERSION && aiRecording) {
			AIRecordingCheckGameOver();
			return;
		}
		
		var facLen = activeFactions.length,
			aliveCount = 0,
			playerAlive = (playerFaction.pixels.length > 0);
		for (var i = 0; i < facLen; i++) {
			var nPx = activeFactions[i].pixels.length;
			gameFactionCounts[i].textContent = nPx+"";
			if (nPx > 0) aliveCount++;
		}
		
		if (playerAlive && aliveCount === 1) GameOver(true,false);//player won
		else if (aliveCount === 0) GameOver(false,true);//tie
		else if (!playerAlive) GameOver(false,false);//player lost
	}
}

