//Ethan Alexander Shulman 2024

var cameraX = 1024, cameraY = 1024, aspectX = 1, aspectY = 1, cameraZoom = 256,
winWidth = 0, winHeight = 0,
bgShader, selectShader, targetShader, pointShader, mineShader, pathShader, pixelShader,
deathFxShader, waveSimShader,
deathFxA = null, deathFxB = null, deathFxQuality = 0, lastDeadPixel = 0, quadIndexBuffer,
SPRITE_ATTRIBS;

const FBUF8 = new Float32Array(8),
FBUF16 = new Float32Array(16),
FBUF32 = new Float32Array(32),
SPRITE_INDICES_LUT = [2,1,0, 3,1,2],

DEATH_EFFECT_OFF = 0,
DEATH_EFFECT_LOW = 1,
DEATH_EFFECT_HIGH = 2;


//initialize rendering systems
function InitializeRendering() {
	SPRITE_ATTRIBS = [new MeshAttrib("vPos",4,4,gl.FLOAT)];
	
	//create one big index buffer of indices for sprite quads
	var ind = new Uint16Array(16384*6), i = 0, fi = 0;
	for (var q = 0; q < 16384; q++) {
		ind[i++] = fi+2;
		ind[i++] = fi+1;
		ind[i++] = fi;
		ind[i++] = fi+3;
		ind[i++] = fi+1;
		ind[i++] = fi+2;
		fi += 4;
	}
	quadIndexBuffer = gl.createBuffer();
	gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, quadIndexBuffer);
	gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, ind, gl.STATIC_DRAW);

	//shaders
	var svert = LoadShader("SpriteVert.glsl", gl.VERTEX_SHADER),
		pointFrag = LoadShader("Point.glsl", gl.FRAGMENT_SHADER),
		mineFrag = LoadShader("Mine.glsl", gl.FRAGMENT_SHADER),
		pathFrag = LoadShader("Path.glsl", gl.FRAGMENT_SHADER),
		pixelFrag = LoadShader("Pixel.glsl", gl.FRAGMENT_SHADER);
	pointShader = new Shader(svert,pointFrag,SPRITE_ATTRIBS);
	mineShader = new Shader(svert,mineFrag,SPRITE_ATTRIBS);
	pathShader = new Shader(svert,pathFrag,SPRITE_ATTRIBS);
	pixelShader = new Shader(svert,pixelFrag,SPRITE_ATTRIBS);
	gl.deleteShader(svert);
	gl.deleteShader(pointFrag);
	gl.deleteShader(mineFrag);
	gl.deleteShader(pathFrag);
	gl.deleteShader(pixelFrag);
	
	bgShader = LoadShaders("BgVert.glsl","Bg.glsl", fullscreenTriangle.attributes);
	selectShader = LoadShaders("SelectVert.glsl","Select.glsl", fullscreenQuad.attributes);
	targetShader = LoadShaders("TargetVert.glsl","Target.glsl", fullscreenQuad.attributes);
	deathFxShader = LoadShaders("DeathEffectsVert.glsl","DeathEffects.glsl", fullscreenTriangle.attributes);
	waveSimShader = LoadShaders("WaveSimVert.glsl","WaveSim.glsl", fullscreenTriangle.attributes);
	

	//window display resizing
	OnWindowResize();
	window.addEventListener("resize", OnWindowResize);
}


//on window resize event
function OnWindowResize() {
	SetDisplaySize(winWidth = window.innerWidth, winHeight = window.innerHeight, true);
	
	//resize aspect
	var len = 1/(display.width+display.height),
		asX = display.width*len,
		asY = display.height*len;
	len = 1/Math.sqrt(asX*asY);
	aspectX = asX*len;
	aspectY = asY*len;
}


//world and screen coordinate conversion
function WorldToScreen(wx,wy) {
	returnY = ((wy-cameraY)/(aspectY*cameraZoom)*-0.5+0.5)*display.height;
	return ((wx-cameraX)/(aspectX*cameraZoom)*0.5+0.5)*display.width;
}
function ScreenToWorld(sx,sy) {
	returnY = (sy/display.height*-2+1)*(aspectY*cameraZoom) + cameraY;
	return (sx/display.width*2-1)*(aspectX*cameraZoom) + cameraX;
}
function ScaleWorldToScreen(wsx,wsy) {
	returnY = wsy/(aspectY*cameraZoom)*display.height*-0.5;
	return wsx/(aspectX*cameraZoom)*display.width*0.5;
}
function ScaleScreenToWorld(ssx,ssy) {
	returnY = (ssy/display.height)*(aspectY*cameraZoom)*-2;
	return (ssx/display.width)*(aspectX*cameraZoom)*2;
}



//sprite batch, multiple sprites sharing the same shader/uniforms rendered together
/**@constructor*/
function SpriteBatch() {
	this.meshes = [];
	this.bufferSize = 0;
	this.renderCount = 0;
	this.verts = null;
	this.indices = null;
	this.lines = false;
}

//update sprite batch
SpriteBatch.prototype.Update = function(sps, staticDraw) {
	//resize buffers
	var nsprites = sps.length, bufSz = this.bufferSize,
		reqMeshes = Math.ceil(nsprites/16384), reqBufSz = (reqMeshes > 1 ? 16384 : nsprites),
		allocMeshes = (this.meshes.length < reqMeshes),
		allocBuf = (bufSz < reqBufSz);
	if (allocMeshes || allocBuf) {
		if (allocBuf) {
			if (bufSz === 0) bufSz = 1;
			while (bufSz < reqBufSz) {
				bufSz *= 2;
			}
			this.verts = new Float32Array(16*bufSz);
			this.bufferSize = bufSz;
		}
		if (allocMeshes) {
			for (var i = this.meshes.length; i < reqMeshes; i++) {
				var msh = new Mesh(SPRITE_ATTRIBS);
				msh.indexBuffer = quadIndexBuffer;
				msh.sharedIndices = true;
				this.meshes.push(msh);
			}
		}
	}
	
	//build mesh vertices
	var lines = this.lines,
		va = this.verts, vi = 0,
		meshes = this.meshes, mi = 0,
		s = 0;
	while (s < nsprites) {
		var iter = Math.min(16384,nsprites-s);
		vi = 0;
		for (var i = 0; i < iter; i++) {
			var sprite = sps[s],
				x = sprite.x, y = sprite.y,
				sx = sprite.scaleX, sy = sprite.scaleY;
			if (lines) {
				var acs = Math.cos(sprite.angle), asn = Math.sin(sprite.angle);
				
				//top left
				va[vi++] = x - sx*acs + sy*asn;
				va[vi++] = y - sx*-asn + sy*acs;
				va[vi++] = -1;
				va[vi++] = 1;
			
				//top right
				va[vi++] = x + sx*acs + sy*asn;
				va[vi++] = y + sx*-asn + sy*acs;
				va[vi++] = 1;
				va[vi++] = 1;
			
				//bottom left
				va[vi++] = x - sx*acs - sy*asn;
				va[vi++] = y - sx*-asn - sy*acs;
				va[vi++] = -1;
				va[vi++] = -1;

				//bottom right
				va[vi++] = x + sx*acs - sy*asn;
				va[vi++] = y + sx*-asn - sy*acs;
				va[vi++] = 1;
				va[vi++] = -1;
			
			} else {
				//top left
				va[vi++] = x - sx;
				va[vi++] = y + sy;
				va[vi++] = -1;
				va[vi++] = 1;
			
				//top right
				va[vi++] = x + sx;
				va[vi++] = y + sy;
				va[vi++] = 1;
				va[vi++] = 1;
			
				//bottom left
				va[vi++] = x - sx;
				va[vi++] = y - sy;
				va[vi++] = -1;
				va[vi++] = -1;

				//bottom right
				va[vi++] = x + sx;
				va[vi++] = y - sy;
				va[vi++] = 1;
				va[vi++] = -1;
			}
			s++;
		}
		meshes[mi++].Set((vi === va.length ? va : va.subarray(0,vi)), staticDraw ? gl.STATIC_DRAW : gl.DYNAMIC_DRAW);
	}
	this.renderCount = nsprites;
}

//render sprite batch meshes
SpriteBatch.prototype.Render = function() {
	var rcount = this.renderCount, mi = 0,
		meshes = this.meshes;
	while (rcount) {
		var iter = Math.min(16384,rcount);
		meshes[mi++].Bind();
		gl.drawElements(gl.TRIANGLES,iter*6,gl.UNSIGNED_SHORT,0);
		rcount -= iter;
	}
}


//sprite rendered via sprite batch
/**@constructor*/
function Sprite() {
	this.x = 0;
	this.y = 0;
	this.scaleX = 1;
	this.scaleY = 1;
	this.angle = 0;
}


//update death fx buffer resolution
function UpdateDeathFxQuality(newQuality) {
	if (newQuality === deathFxQuality) return;
	
	var newSz = newQuality*512;
	if (deathFxA) {
		//resize existing buffer
		deathFxA.SetSize(newSz,newSz);
		deathFxB.SetSize(newSz,newSz);
	} else {
		//create buffers
		deathFxA = new RenderTexture(newSz,newSz);
		SetTextureOptions(true,false,false,false);
		deathFxB = new RenderTexture(newSz,newSz);
		SetTextureOptions(true,false,false,false);
	}
	deathFxQuality = newQuality;
}


//render dead pixels into fx simulation
function RenderDeadPixelsFx() {
	deathFxA.Bind();
	pixelShader.Bind();
	var sc = 1/1024;
	gl.uniform4f(pixelShader["camera"], 1024,1024,sc,sc);
	gl.uniform4f(pixelShader["color"], 0.2, 0,0,0);
	selectedBatch.Update(deadPixels,false);
	selectedBatch.Render();
}

//update death fx wave simulation
function UpdateDeathFx() {
	var tmp = deathFxA;
	deathFxA = deathFxB;
	deathFxB = tmp;

	gl.disable(gl.BLEND);
	waveSimShader.Bind();
	deathFxA.Bind();
	fullscreenTriangle.Bind();
	gl.uniform1f(waveSimShader["ext"], 1/deathFxB.width);
	BindTexture(waveSimShader["sim"], deathFxB.texture, null);
	gl.drawArrays(gl.TRIANGLES,0,3);
	UnbindAllTextures();
	gl.enable(gl.BLEND);
}


//render
function Render() {
	var camSX = 1/(aspectX*cameraZoom),
		camSY = 1/(aspectY*cameraZoom);
	display.Bind();
	
	//background
	gl.disable(gl.BLEND);
	bgShader.Bind();
	fullscreenTriangle.Bind();
	gl.uniform4f(bgShader["camera"], cameraX,cameraY,camSX,camSY);
	gl.drawArrays(gl.TRIANGLES,0,3);
	gl.enable(gl.BLEND);
	
	//death effect
	if (deathFxQuality) {
		deathFxShader.Bind();
		fullscreenTriangle.Bind();
		gl.uniform4f(deathFxShader["camera"], cameraX,cameraY,camSX,camSY);
		BindTexture(deathFxShader["sim"], deathFxA.texture, null);
		gl.drawArrays(gl.TRIANGLES,0,3);
		UnbindAllTextures();
	}
	
	//render map
	if (mapPointsBatch.renderCount) {
		pointShader.Bind();
		gl.uniform4f(pointShader["camera"], cameraX,cameraY,camSX,camSY);
		mapPointsBatch.Render();
	}
	if (mapMinesBatch.renderCount) {
		mineShader.Bind();
		gl.uniform4f(mineShader["camera"], cameraX,cameraY,camSX,camSY);
		mapMinesBatch.Render();
	}
	if (mapPathsBatch.renderCount) {
		pathShader.Bind();
		gl.uniform4f(pathShader["camera"], cameraX,cameraY,camSX,camSY);
		mapPathsBatch.Render();
	}
	
	//render faction pixels
	var initPxRender = true,
		ipB = (tickTimer/TICK_STEP)%1, ipA = 1-ipB;
	for (var f = 0; f < activeFactions.length; f++) {
		var fac = activeFactions[f],
			fpx = fac.pixels,
			pxLen = fpx.length;
		if (pxLen) {
			//update interpolated position
			for (var i = 0; i < pxLen; i++) {
				var px = fpx[i];
				px.x = px.lpX*ipA + px.pX*ipB;
				px.y = px.lpY*ipA + px.pY*ipB;
			}
		
			if (initPxRender) {
				pixelShader.Bind();
				gl.uniform4f(pixelShader["camera"], cameraX,cameraY,camSX,camSY);
				initPxRender = false;
			}
			
			var batch = fac.pixelsBatch;
			batch.Update(fpx, false);
			gl.uniform4f(pixelShader["color"], fac.red*0.45,fac.green*0.45,fac.blue*0.45,0);
			batch.Render();
		}
	}
	
	//render selected pixels
	if (selectedPixels.length) {
		selectedBatch.Update(selectedPixels,false);
		gl.uniform4f(pixelShader["color"], 0.3,0.3,0.3,0);
		selectedBatch.Render();
	
		if (targetPoint) {
			targetShader.Bind();
			fullscreenQuad.Bind();
			var b = FBUF8;
			b[0] = cameraX; b[1] = cameraY;
			b[2] = camSX; b[3] = camSY;
			b[4] = targetPoint.x; b[5] = targetPoint.y;
			b[6] = 100; b[7] = 100;
			gl.uniform4fv(targetShader["data"], b);
			gl.drawArrays(gl.TRIANGLE_STRIP,0,4);
		}
	}
	
	//selection box
	if (selectingPixels) {
		selectShader.Bind();
		fullscreenQuad.Bind();
		var b = FBUF8;
		b[0] = cameraX; b[1] = cameraY;
		b[2] = camSX; b[3] = camSY;
		b[4] = selectCenterX; b[5] = selectCenterY;
		b[6] = selectExtentX; b[7] = selectExtentY;
		gl.uniform4fv(selectShader["data"], b);
		gl.drawArrays(gl.TRIANGLE_STRIP,0,4);
	}
}


